#include <TFT_eSPI.h>
#include <Arduino.h>

#ifndef WIO_KEY_A
  #define WIO_KEY_A 6   // Moves tetromino left / accelerates drop: short press left, long press for instant drop
#endif
#ifndef WIO_KEY_B
  #define WIO_KEY_B 5   // Moves tetromino right / restarts game: short press right, long press to restart
#endif
#ifndef WIO_KEY_C
  #define WIO_KEY_C 9   // Rotates tetromino clockwise
#endif

TFT_eSPI tft = TFT_eSPI();

//////////////////////////////
// Game Settings
//////////////////////////////

const int BOARD_WIDTH = 27;     // 27 columns
const int BOARD_HEIGHT = 17;    // 20 rows
const int CELL_SIZE = 12;       // Each cell is 12x12 pixels
const int BOARD_PIXEL_WIDTH = BOARD_WIDTH * CELL_SIZE;   // 324 pixels wide
const int BOARD_PIXEL_HEIGHT = BOARD_HEIGHT * CELL_SIZE; // 240 pixels tall
const int SCORE_AREA_HEIGHT = 30;  // Score display area height

unsigned long dropInterval = 500;  // Tetromino auto-drop interval in ms
unsigned long lastDropTime = 0;

int board[BOARD_HEIGHT][BOARD_WIDTH] = {0};  // Game board array (0 = empty)

int currentPiece;    // Tetromino index (0 to 6)
int currentRotation; // Rotation state (0 to 3)
int pieceX;          // Current tetromino X position (grid units)
int pieceY;          // Current tetromino Y position (grid units)

bool gameOver = false;
int score = 0;
int lastScore = -1;  // Used for determining if the score display needs to be updated

// Key press detection variables for KEY_A and KEY_B:
unsigned long keyAPressStart = 0;
bool keyAProcessed = false;
const unsigned long longPressThreshold = 300; // ms

unsigned long keyBPressStart = 0;
bool keyBProcessed = false;

//////////////////////////////
// Tetromino Definitions
//////////////////////////////
// 7 tetrominoes: I, O, T, S, Z, J, L. Each is defined as a 4x4 matrix for 4 rotation states.
const byte tetrominoes[7][4][4][4] = {
  // I Piece
  {
    { {0,0,0,0}, {1,1,1,1}, {0,0,0,0}, {0,0,0,0} },
    { {0,0,1,0}, {0,0,1,0}, {0,0,1,0}, {0,0,1,0} },
    { {0,0,0,0}, {1,1,1,1}, {0,0,0,0}, {0,0,0,0} },
    { {0,0,1,0}, {0,0,1,0}, {0,0,1,0}, {0,0,1,0} }
  },
  // O Piece
  {
    { {0,1,1,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,1,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,1,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,1,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} }
  },
  // T Piece
  {
    { {0,1,0,0}, {1,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {0,1,1,0}, {0,1,0,0}, {0,0,0,0} },
    { {0,0,0,0}, {1,1,1,0}, {0,1,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {1,1,0,0}, {0,1,0,0}, {0,0,0,0} }
  },
  // S Piece
  {
    { {0,1,1,0}, {1,1,0,0}, {0,0,0,0}, {0,0,0,0} },
    { {1,0,0,0}, {1,1,0,0}, {0,1,0,0}, {0,0,0,0} },
    { {0,1,1,0}, {1,1,0,0}, {0,0,0,0}, {0,0,0,0} },
    { {1,0,0,0}, {1,1,0,0}, {0,1,0,0}, {0,0,0,0} }
  },
  // Z Piece
  {
    { {1,1,0,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {1,1,0,0}, {1,0,0,0}, {0,0,0,0} },
    { {1,1,0,0}, {0,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {1,1,0,0}, {1,0,0,0}, {0,0,0,0} }
  },
  // J Piece
  {
    { {1,0,0,0}, {1,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,1,0}, {0,1,0,0}, {0,1,0,0}, {0,0,0,0} },
    { {1,1,1,0}, {0,0,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {0,1,0,0}, {1,1,0,0}, {0,0,0,0} }
  },
  // L Piece
  {
    { {0,0,1,0}, {1,1,1,0}, {0,0,0,0}, {0,0,0,0} },
    { {0,1,0,0}, {0,1,0,0}, {0,1,1,0}, {0,0,0,0} },
    { {1,1,1,0}, {1,0,0,0}, {0,0,0,0}, {0,0,0,0} },
    { {1,1,0,0}, {0,1,0,0}, {0,1,0,0}, {0,0,0,0} }
  }
};

//////////////////////////////
// Function Declarations
//////////////////////////////
bool isValidPosition(int newX, int newY, int newRotation);
void spawnPiece();
void placePiece();
void clearLines();
void drawBoard();
void restartGame();
void drawScore();

//////////////////////////////
// Function Implementations
//////////////////////////////

// Checks if placing the tetromino at a new position with a given rotation is valid
bool isValidPosition(int newX, int newY, int newRotation) {
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      byte cell = tetrominoes[currentPiece][newRotation][i][j];
      if (cell) {
        int boardX = newX + j;
        int boardY = newY + i;
        if (boardX < 0 || boardX >= BOARD_WIDTH || boardY < 0 || boardY >= BOARD_HEIGHT)
          return false;
        if (board[boardY][boardX] != 0)
          return false;
      }
    }
  }
  return true;
}

// Spawns a new random tetromino and resets its position and rotation
void spawnPiece() {
  currentPiece = random(0, 7);  // Random index from 0 to 6
  currentRotation = 0;
  pieceX = (BOARD_WIDTH - 4) / 2;
  pieceY = 0;
  if (!isValidPosition(pieceX, pieceY, currentRotation))
    gameOver = true;
}

// Fixes the tetromino on the board, clears lines, and spawns a new piece
void placePiece() {
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      if (tetrominoes[currentPiece][currentRotation][i][j]) {
        int boardX = pieceX + j;
        int boardY = pieceY + i;
        if (boardX >= 0 && boardX < BOARD_WIDTH && boardY >= 0 && boardY < BOARD_HEIGHT)
          board[boardY][boardX] = currentPiece + 1;
      }
    }
  }
  clearLines();
  spawnPiece();
}

// Checks for full lines on the board, clears them, increments score, and updates the score display.
void clearLines() {
  for (int i = 0; i < BOARD_HEIGHT; i++) {
    bool full = true;
    for (int j = 0; j < BOARD_WIDTH; j++) {
      if (board[i][j] == 0) {
        full = false;
        break;
      }
    }
    if (full) {
      score++;  // Increase score by 1 per cleared line
      for (int k = i; k > 0; k--) {
        for (int j = 0; j < BOARD_WIDTH; j++) {
          board[k][j] = board[k - 1][j];
        }
      }
      for (int j = 0; j < BOARD_WIDTH; j++)
        board[0][j] = 0;
      // Update the score display when a line is cleared.
      drawScore();
    }
  }
}

// Draws the score in the dedicated score area.
void drawScore() {
  tft.startWrite();
  tft.fillRect(0, 0, BOARD_PIXEL_WIDTH, SCORE_AREA_HEIGHT, TFT_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.setCursor(10, 10);
  tft.print("Score: ");
  tft.print(score);
  tft.endWrite();
  lastScore = score;
}

// Draws the complete board, including fixed blocks, the falling tetromino, and game over message if needed.
void drawBoard() {
  tft.startWrite();
  
  // Draw the board area (starting below the score area)
  tft.fillRect(0, SCORE_AREA_HEIGHT, BOARD_PIXEL_WIDTH, BOARD_PIXEL_HEIGHT, TFT_BLACK);
  
  // Draw fixed blocks on the board
  for (int i = 0; i < BOARD_HEIGHT; i++) {
    for (int j = 0; j < BOARD_WIDTH; j++) {
      if (board[i][j] != 0) {
        uint16_t color;
        switch(board[i][j]) {
          case 1: color = TFT_CYAN; break;
          case 2: color = TFT_YELLOW; break;
          case 3: color = TFT_MAGENTA; break;
          case 4: color = TFT_GREEN; break;
          case 5: color = TFT_RED; break;
          case 6: color = TFT_BLUE; break;
          case 7: color = TFT_ORANGE; break;
          default: color = TFT_WHITE; break;
        }
        tft.fillRect(j * CELL_SIZE, i * CELL_SIZE + SCORE_AREA_HEIGHT, CELL_SIZE - 1, CELL_SIZE - 1, color);
      }
    }
  }
  
  // Draw the current falling tetromino
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      if (tetrominoes[currentPiece][currentRotation][i][j]) {
        int boardX = pieceX + j;
        int boardY = pieceY + i;
        if (boardX >= 0 && boardX < BOARD_WIDTH && boardY >= 0 && boardY < BOARD_HEIGHT) {
          uint16_t color;
          switch(currentPiece) {
            case 0: color = TFT_CYAN; break;
            case 1: color = TFT_YELLOW; break;
            case 2: color = TFT_MAGENTA; break;
            case 3: color = TFT_GREEN; break;
            case 4: color = TFT_RED; break;
            case 5: color = TFT_BLUE; break;
            case 6: color = TFT_ORANGE; break;
            default: color = TFT_WHITE; break;
          }
          tft.fillRect(boardX * CELL_SIZE, boardY * CELL_SIZE + SCORE_AREA_HEIGHT, CELL_SIZE - 1, CELL_SIZE - 1, color);
        }
      }
    }
  }
  
  // If the game is over, display "Game Over" in the center
  if (gameOver) {
    tft.setTextSize(3);
    tft.setTextColor(TFT_RED, TFT_BLACK);
    int16_t x = (BOARD_PIXEL_WIDTH - 6 * 12) / 2;
    int16_t y = (BOARD_PIXEL_HEIGHT - 24) / 2 + SCORE_AREA_HEIGHT;
    tft.setCursor(x, y);
    tft.print("Game Over");
  }
  
  tft.endWrite();
}

// Restarts the game by clearing the board, resetting score and game variables, then spawning a new piece.
void restartGame() {
  for (int i = 0; i < BOARD_HEIGHT; i++)
    for (int j = 0; j < BOARD_WIDTH; j++)
      board[i][j] = 0;
  score = 0;
  lastScore = -1;
  gameOver = false;
  spawnPiece();
  lastDropTime = millis();
  drawScore();
}

//////////////////////////////
// Setup and Loop
//////////////////////////////

void setup() {
  Serial.begin(115200);
  tft.init();
  tft.setRotation(1);
  tft.fillScreen(TFT_BLACK);
  
  pinMode(WIO_KEY_A, INPUT_PULLUP);
  pinMode(WIO_KEY_B, INPUT_PULLUP);
  pinMode(WIO_KEY_C, INPUT_PULLUP);
  
  // Initialize board to zero
  for (int i = 0; i < BOARD_HEIGHT; i++)
    for (int j = 0; j < BOARD_WIDTH; j++)
      board[i][j] = 0;
      
  randomSeed(analogRead(0));
  spawnPiece();
  lastDropTime = millis();
  drawScore();
}

void loop() {
  unsigned long now = millis();
  
  // When game is over, only process KEY_B for restart.
  if (gameOver) {
    if (digitalRead(WIO_KEY_B) == LOW) {
      if (keyBPressStart == 0) {
        keyBPressStart = millis();
        keyBProcessed = false;
      } else {
        if (millis() - keyBPressStart >= longPressThreshold && !keyBProcessed) {
          restartGame();
          keyBProcessed = true;
        }
      }
    } else {
      keyBPressStart = 0;
      keyBProcessed = false;
    }
    drawBoard();
    delay(30);
    return;
  }
  
  // Process KEY_A: Short press moves left; long press (>300ms) accelerates drop.
  if (digitalRead(WIO_KEY_A) == LOW) {
    if (keyAPressStart == 0) {
      keyAPressStart = millis();
      keyAProcessed = false;
    } else {
      if (millis() - keyAPressStart >= longPressThreshold && !keyAProcessed) {
        while (isValidPosition(pieceX, pieceY + 1, currentRotation))
          pieceY++;
        placePiece();
        keyAProcessed = true;
      }
    }
  } else {
    if (keyAPressStart != 0 && !keyAProcessed) {
      if (isValidPosition(pieceX - 1, pieceY, currentRotation))
        pieceX--;
    }
    keyAPressStart = 0;
    keyAProcessed = false;
  }
  
  // Process KEY_B: Short press moves right; long press (>300ms) restarts the game.
  if (digitalRead(WIO_KEY_B) == LOW) {
    if (keyBPressStart == 0) {
      keyBPressStart = millis();
      keyBProcessed = false;
    } else {
      if (millis() - keyBPressStart >= longPressThreshold && !keyBProcessed) {
        restartGame();
        keyBProcessed = true;
      }
    }
  } else {
    if (keyBPressStart != 0 && !keyBProcessed) {
      if (isValidPosition(pieceX + 1, pieceY, currentRotation))
        pieceX++;
    }
    keyBPressStart = 0;
    keyBProcessed = false;
  }
  
  // Process KEY_C: Rotate tetromino clockwise.
  if (digitalRead(WIO_KEY_C) == LOW) {
    int newRot = (currentRotation + 1) % 4;
    if (isValidPosition(pieceX, pieceY, newRot))
      currentRotation = newRot;
    delay(50);
  }
  
  // Automatic drop of tetromino
  if (now - lastDropTime > dropInterval) {
    if (isValidPosition(pieceX, pieceY + 1, currentRotation))
      pieceY++;
    else
      placePiece();
    lastDropTime = now;
  }
  
  drawBoard();
  delay(30);
}

